#include "converterfunctions.h"
#include "rrob_to_gcol.h"

void convertToGCol(int n, bool **adjacent, int numMatchNodes, bool doubleRR, bool outputToScreen, 
				   int *hometeam, int *awayteam)
{
	//THIS PROCEDURE GOES THRU THE ADJ. MATRIX AND CONVERTS THE RELEVANT ELEMENTS TO FALSE + TRUE
	//I.E. EDGES ARE GEING ADDED
	int r,x,c,y,j,i,len;

	//First, make all entries in the adjaceny matrix false
	for(r=0; r<numMatchNodes; r++) for(c=0; c<numMatchNodes; c++) adjacent[r][c]=false;

	//Now make all nodes adjacent to themselves
	for(r=0; r<numMatchNodes; r++) adjacent[r][r]=true;

	//Now work out which nodes correspond to which matches and store these all in two arrays
	if(outputToScreen){
		cout<<"The nodes correspond to the following matches ";
		if(doubleRR)cout<<"(home,away): [nb: Ordering is important]\n";
		else cout<<"{t_1,t_2}: [Ordering is unimportant]\n";
	}


	//-----------------------------------------------------------------------------
	//NOW ASSOCIATE EACH NODE WITH A DISTINCT PAIR OF TEAMS
	if(doubleRR) len = numMatchNodes/2;
	else len = numMatchNodes;

	x = 0;
	y = x+1;
	for(i=0; i<len; i++){
		
		if(outputToScreen){
			if(doubleRR) cout<<i<<" = ("<<x<<","<<y<<")\n";
			else cout<<i<<" = {"<<x<<","<<y<<"}\n";
		}

		hometeam[i] = x;
		awayteam[i] = y;
		
		//Fill the home and away team arrays according to the single RR specification
		if(y==n-1){
			x++;
			y=x+1;
		}
		else y++;
	}

	if(doubleRR){
		//IF THE PROBLEM IS A DOUBLE RR, we now fill in the second half of the home and away matches,
		//-- this is simply a copy of the first half with the fixtures reversed.
		for(i=len; i<numMatchNodes; i++){

			hometeam[i] = awayteam[i-len];
			awayteam[i] = hometeam[i-len];

			if(outputToScreen)cout<<i<<" = ("<<hometeam[i-len]<<","<<awayteam[i-len]<<")\n";
		}
	}


	//Now add edges between any node pairs that represent a match with a common team
	//Nb: each node i = 0,...,n-1 represents a pair of teams (x,y) who are playing each other
	//The pairs are arranged as follows for single RR {(0,1),(0,2),....(0,n-1),(1,2),......(n-2,n-1)}
	//and {(0,1),(0,2),....(0,n-1),(1,2),(1,3)... etc in the first half, and then (1,0), (2,0) etc} for double RR
	for(i=0; i<numMatchNodes-1; i++){
		for(j=i+1; j<numMatchNodes; j++){
			if(hometeam[i] == hometeam[j] 
				|| hometeam[i] == awayteam[j]
				|| awayteam[i] == hometeam[j]
				|| awayteam[i] == awayteam[j]){
				adjacent[i][j] = true;
				adjacent[j][i] = true;
			}
		}
	}
}



//------------------------------------------------------------------------------------------------
int calcNumEdges(bool **adjacent, int numNodes)
{
	int r, c;
	int count = 0;
	for(r=0; r<numNodes; r++){
		for(c=0; c<r; c++){
			if(adjacent[r][c])count++;
		}
	}
	return count;
}
	


//------------------------------------------------------------------------------------------------
void makeOutput(char *filename, ofstream &outStream, int numMatchNodes, bool **adjacent, int numMatchEdges, int n,
				bool doubleRR, int *hometeam, int *awayteam, bool doingOtherConstraints, bool isMirrored, 
				bool perfectSolution, double preAssignProb, double unavailProb, char *perfectFilename,
				int rSeed, bool writeComments, bool outputToScreen)
{
	//PRINT OUT AL THE DETAILS OF THE GRAPH COLOURING PROBLEM TO THE OUTPUT FILE. IN THIS PART WE ALSO
	//ADD THE DETAILS OF THE ROUND NODES IF NECESSARY
	int r, c, i, numTSNodes, numNodes, j, R; 
	
	//Let R = the number of rounds in the tournament
	if(doubleRR) R = (n-1)*2;
	else R = n-1;

	if(doingOtherConstraints){
		numTSNodes = R;
		numNodes = numMatchNodes + numTSNodes;
	}
	else numNodes = numMatchNodes; 


	//DO SOME OUTPUT STUFF
	if(writeComments){
		outStream<<"c FILE "<<filename<<endl;
		outStream<<"c\n";
		outStream<<"c\n";
		outStream<<"c SOURCE This file is a graph colouring problem that arises when trying"<<endl;
		outStream<<"c        to produce a ";
		if(doubleRR){
			if(isMirrored)
				outStream<<"SYMMETRICAL";
			else outStream<<"DOUBLE";
		}
		else outStream<<"SINGLE";
		outStream<<" round robin tournament with "<<n<<" teams";
		if(doingOtherConstraints)
			outStream<<" and "<<numTSNodes<<" timeslot nodes\n";
		else outStream<<"\n";	
		outStream<<"c\n";
		outStream<<"c\n";
		outStream<<"c DESCRIPTION This graph colouring problem was generated \n";
		outStream<<"c    using the converter designed by Rhydian Lewis\n";
		outStream<<"c    If you notice any problems with this program then please contact\n";
		outStream<<"c    Rhydian Lewis by emailing LewisR9@cf.ac.uk\n";
		outStream<<"c\n";
		outStream<<"c\n";
		outStream<<"c GRAPH STATS\n";
		outStream<<"c Number of Match Nodes                       = "<<numMatchNodes<<endl;
		outStream<<"c Number of colours needed for valid RR       = "<<R<<endl;
		outStream<<"c\n";
		if(doingOtherConstraints){
			outStream<<"c There are also "<<numTSNodes<<" additional nodes, indexed "<<numMatchNodes+1<<" to "<<numMatchNodes+numTSNodes<<"\n";
			outStream<<"c These correspond to the "<<numTSNodes<<" timeslots that the matches (match nodes) need to be assigned to."<<endl;
			outStream<<"c A propotion of "<< preAssignProb <<" matches are preassigned to a round\n";
			outStream<<"c A propotion of "<< unavailProb <<" match/timeslot pairs have been assigned as unavailable\n";
			if(perfectSolution)
				outStream<<"c A perfect soltion is guaranteed, however\n";
			else
				outStream<<"c NB: there is no gurantee of a valid round robin in this instance\n";
			outStream<<"c The random seed used was "<<rSeed<<endl;
		}
		else outStream<<"c THERE WILL BE A VALID RR TOURNAMENT IN THIS CASE"<<endl;
		if(doubleRR){
			outStream<<"c \nc THE following nodes correspond to the following (home,away) matches:\nc (note: each team plays each other twice)\n";
			for(i=0; i<numMatchNodes; i++) outStream<<"c node "<<i+1<<"\t=("<<hometeam[i]<<","<<awayteam[i]<<")\n";
		}		
		else {
			outStream<<"c \nc THE following nodes correspond to the following matches:\nc (note: as this is a single RR the ordering in each pair is unimportant here)\n";
			for(i=0; i<numMatchNodes; i++) outStream<<"c node "<<i+1<<"\t={"<<hometeam[i]<<","<<awayteam[i]<<"}\n";
		}
		outStream<<"c \n";
		if(doingOtherConstraints){
			for(i=numMatchNodes+1; i<numMatchNodes+numTSNodes+1; i++) 
				outStream<<"c node "<<i<<"\t= timeslot "<<i-numMatchNodes<<"\n";
		}
	}

	IntVector TS, M;
	if(doingOtherConstraints){
		//If extra constraints are being considered these are created here. The idea here is to 
		//form 2 equi-sized 1D vectors that specify the extra edges occuring.
		if(perfectSolution){
			//Constraints are added so that a perfect solution is guaranteed.
			addOtherConstraintsPerfect(numMatchNodes,numTSNodes,preAssignProb,unavailProb,TS,M,outStream,perfectFilename,writeComments,outputToScreen);
		}
		else
			addOtherConstraintsRandomly(numMatchNodes,numTSNodes,preAssignProb,unavailProb,TS,M,outStream,writeComments);
	}

	//We now want to write out the remainder of the file (i.e. the edges). If approproate, the 
	//extra constraints are added here
	outStream<<"p edge "<<numNodes;
	if(doingOtherConstraints){
		
		//need to work out the number of edges (the rightmost bit of this fla are the edges caused by the clique
		//formed by the timeslot nodes)
		long ne = numMatchEdges + M.size() + ( (numTSNodes*(numTSNodes-1)) /2);
		outStream <<" "<<ne<<endl;
		
		//Now write out the match edges (i.e. the normal edges between match nodes)
		for(r=0; r<numMatchNodes; r++){
			for(c=0; c<r; c++){
				if(adjacent[r][c])outStream<<"e "<<r+1<<' '<<c+1<<'\n';
			}
		}
		
		//Now write out the timeslotNodes clique
		for(i=numMatchNodes+1; i<(numMatchNodes+1)+numTSNodes; i++){
			for(j=i+1; j<(numMatchNodes+1)+numTSNodes; j++){
				outStream<<"e "<<i<<' '<<j<<'\n';
			}
		}
		
		//Now add the edges due to the imposed "other" constraints
		for(i=0;i<M.size();i++){
			outStream<<"e "<<M[i]<<' '<<TS[i]<<'\n';
		}
	}
	else {
		//Simply write out the edges occuring between the match nodes
		outStream<<" "<<numMatchEdges<<endl;
		for(r=0; r<numMatchNodes; r++){
			for(c=0; c<r; c++){
				if(adjacent[r][c])outStream<<"e "<<r+1<<' '<<c+1<<'\n';
			}
		}
	}

	cout<<"Instance created succesfully\n";

}

//***********************************************************************************
void addOtherConstraintsRandomly(int numMatchNodes, int numTSNodes, double preAssignProb, double unavailProb, IntVector &TS,
						 IntVector &M, ofstream &outStream, bool writeComments)
{
	int j, i, k;
	double p;

	if(writeComments)
		outStream<<"c \nc The Pre-assignments are as follows: "<<endl;
	//Go thru each match. If we decide to reassign it, select a slot at random and preassign it here
	for(i=1;i<numMatchNodes+1;i++){
		p = (rand()%1024)/1024.0;
		if(p<preAssignProb){
			//Select a round at random
			j = rand()%numTSNodes;
			j = j + (numMatchNodes+1);

			//Match i is to be preassigned to round j
			if(writeComments)
				outStream<<"c Match "<<i<<" is pre-assigned to timeslot "<<j-numMatchNodes<<endl;
		
			//Now add edges between match i and all timeslots except j
			for(k=numMatchNodes+1; k<numMatchNodes+numTSNodes+1; k++){
				if(k != j){
					M.push_back(i);
					TS.push_back(k);
				}
			}
		}
	}
	if(writeComments) 
		outStream<<"c \nc The match unavailabilities are as follows: "<<endl;
	//Determine which TSNodes and Match nodes are to have edges between them
	for(i=1;i<numMatchNodes+1;i++){
		for(j=numMatchNodes+1;j<numMatchNodes+numTSNodes+1;j++){
			p = (rand()%1024)/1024.0;
			if(p<unavailProb){
				//we add an edge between event i and timeslot j
				M.push_back(i);
				TS.push_back(j);
				if(writeComments)
					outStream<<"c Match "<<i<<" unavailable in timeslot "<<j-numMatchNodes<<endl;
			}
		}
	}
}



//***********************************************************************************
void addOtherConstraintsPerfect(int numMatchNodes, int numTSNodes, double preAssignProb, double unavailProb, IntVector &TS,
						 IntVector &M, ofstream &outStream, char *perfectFilename, bool writeComments, bool outputToScreen)
{
	int j, i, k, x, V;
	double p;

	//Read in the perfect solution
	IntVector perfectSolution;
	ifstream inStr;
	inStr.open(perfectFilename);
	if(inStr.fail()) { cout << "Error opening perfect input file. Exiting...\n"; exit(1);}
	inStr >> V;
	if(V!=numMatchNodes){
		cout << "Input solution is a different size to the current problem. Exiting...\n"; exit(1);}
	for(i=0; i<numMatchNodes; i++){
		inStr >> x;
		//Remember that nodes (and groups) are indexed from 1 upwards here
		perfectSolution.push_back(x+1);
	}

	if(outputToScreen){
		for(i=0; i<perfectSolution.size(); i++) cout<<"Node "<<i+1<<" = group "<<perfectSolution[i]<<endl;
	}

	//We now need to adjust the probabilities slightly. This is because the addition of some edges is disaallowed
	//because we want to preserve the perfect property. Note we only need to do this for unavailProbabiliy
	//Considering unavailProb, the number of edges (constraints) we can expect to add is: 
	double numC = unavailProb*(numMatchNodes*numTSNodes);
	//However, because the addition of some constraints is disallowed, the number of possible cases considered is:
	double numCases = (numMatchNodes*numTSNodes) - numMatchNodes;
	//Thus, the probability is:
	double adjUnAvP = numC/numCases;

	if(writeComments) outStream<<"c \nc The match unavailabilities are as follows: "<<endl;
	
	//Determine which TSNodes and Match nodes are to have edges between them
	for(i=1;i<numMatchNodes+1;i++){
		for(j=numMatchNodes+1;j<numMatchNodes+numTSNodes+1;j++){
			p = (rand()%1024)/1024.0;
			if(p<adjUnAvP &&    ( perfectSolution[i-1] != (j-numMatchNodes) )    ){
				//i.e. if probability dictates, AND we know that node i is not assigned to group j in the perfect
				//soltuon, we can add an edge between (match)node i and (timeslot)node j
				M.push_back(i);
				TS.push_back(j);
				if(writeComments)
					outStream<<"c Match "<<i<<" unavailable in timeslot "<<j-numMatchNodes<<endl;
			}
		}
	}

	if(writeComments) outStream<<"c \nc The Pre-assignments are as follows: "<<endl;

	//Go thru each match. If we decide to reassign it, select a slot at random and preassign it here
	for(i=1;i<numMatchNodes+1;i++){
		p = (rand()%1024)/1024.0;
		if(p<preAssignProb){
			
			//if probability dictates, match i should be preassigned to the slot it is in in the perfect solution
			j = perfectSolution[i-1];
			j = j + numMatchNodes;
			
			//Match i is to be preassigned to round j
			if(writeComments)
				outStream<<"c Match "<<i<<" is pre-assigned to timeslot "<<j-numMatchNodes<<endl;
			
			//Now add edges between match i and all timeslots except j
			for(k=numMatchNodes+1; k<numMatchNodes+numTSNodes+1; k++){
				if(k != j){
					M.push_back(i);
					TS.push_back(k);
				}
			}
		}
	}
}


